/** \file
 *  \brief Declares the Event class and helpers
 *
 *  \authors   Jeppe Andersen, Tuomas Hapola, Marian Heil, Andreas Maier, Jennifer Smillie
 *  \date      2019
 *  \copyright GPLv2 or later
 */

#pragma once

#include <array>
#include <memory>
#include <string>
#include <unordered_map>
#include <vector>

#include "HEJ/event_types.hh"
#include "HEJ/Particle.hh"

#include "fastjet/ClusterSequence.hh"

namespace LHEF{
  class HEPEUP;
  class HEPRUP;
}

namespace fastjet{
  class JetDefinition;
}

namespace HEJ{


  struct ParameterDescription;

  //! Event parameters
  struct EventParameters{
    double mur;              /**< Value of the Renormalisation Scale */
    double muf;              /**< Value of the Factorisation Scale */
    double weight;           /**< Event Weight */
    //! Optional description
    std::shared_ptr<ParameterDescription> description = nullptr;
  };

  //! Description of event parameters
  struct ParameterDescription {
    //! Name of central scale choice (e.g. "H_T/2")
    std::string scale_name;
    //! Actual renormalisation scale divided by central scale
    double mur_factor;
    //! Actual factorisation scale divided by central scale
    double muf_factor;

    ParameterDescription() = default;
    ParameterDescription(
        std::string scale_name, double mur_factor, double muf_factor
    ):
      scale_name{scale_name}, mur_factor{mur_factor}, muf_factor{muf_factor}
    {};
  };

  //! An event before jet clustering
  struct UnclusteredEvent{
    //! Default Constructor
    UnclusteredEvent() = default;
    //! Constructor from LesHouches event information
    UnclusteredEvent(LHEF::HEPEUP const & hepeup);

    std::array<Particle, 2> incoming;          /**< Incoming Particles */
    std::vector<Particle> outgoing;            /**< Outgoing Particles */
    //! Particle decays in the format {outgoing index, decay products}
    std::unordered_map<size_t, std::vector<Particle>> decays;
    //! Central parameter (e.g. scale) choice
    EventParameters central;
    std::vector<EventParameters> variations;    /**< For parameter variation */
  };

  /** An event with clustered jets
    *
    * This is the main HEJ 2 event class.
    * It contains kinematic information including jet clustering,
    * parameter (e.g. scale) settings and the event weight.
    */
  class Event{
  public:
    //! Default Event Constructor
    Event() = default;
    //! Event Constructor adding jet clustering to an unclustered event
    Event(
        UnclusteredEvent ev,
        fastjet::JetDefinition const & jet_def, double min_jet_pt
    );

    //! The jets formed by the outgoing partons
    std::vector<fastjet::PseudoJet> jets() const;

    //! The corresponding event before jet clustering
    UnclusteredEvent const & unclustered() const {
      return ev_;
    }

    //! Central parameter choice
    EventParameters const & central() const{
      return ev_.central;
    }

    //! Central parameter choice
    EventParameters & central(){
      return ev_.central;
    }

    //! Incoming particles
    std::array<Particle, 2> const &  incoming() const{
      return ev_.incoming;
    }

    //! Outgoing particles
    std::vector<Particle> const &  outgoing() const{
      return ev_.outgoing;
    }

    //! Particle decays
    /**
     *  The key in the returned map corresponds to the index in the
     *  vector returned by outgoing()
     */
    std::unordered_map<size_t, std::vector<Particle>> const &  decays() const{
      return ev_.decays;
    }

    //! Parameter (scale) variations
    std::vector<EventParameters> const & variations() const{
      return ev_.variations;
    }

    //! Parameter (scale) variations
    std::vector<EventParameters> & variations(){
      return ev_.variations;
    }

    //! Parameter (scale) variation
    /**
     *  @param i   Index of the requested variation
     */
    EventParameters const & variations(size_t i) const{
      return ev_.variations[i];
    }

    //! Parameter (scale) variation
    /**
     *  @param i   Index of the requested variation
     */
    EventParameters & variations(size_t i){
      return ev_.variations[i];
    }

    //! Indices of the jets the outgoing partons belong to
    /**
     *  @param jets   Jets to be tested
     *  @returns      A vector containing, for each outgoing parton,
     *                the index in the vector of jets the considered parton
     *                belongs to. If the parton is not inside any of the
     *                passed jets, the corresponding index is set to -1.
     */
    std::vector<int> particle_jet_indices(
        std::vector<fastjet::PseudoJet> const & jets
    ) const{
      return cs_.particle_jet_indices(jets);
    }

    //! Jet definition used for clustering
    fastjet::JetDefinition const & jet_def() const{
      return cs_.jet_def();
    }

    //! Minimum jet transverse momentum
    double min_jet_pt() const{
      return min_jet_pt_;
    }

    //! Event type
    event_type::EventType type() const{
      return type_;
    }

  private:
    UnclusteredEvent ev_;
    fastjet::ClusterSequence cs_;
    double min_jet_pt_;
    event_type::EventType type_;
  };

  //! Square of the partonic centre-of-mass energy \f$\hat{s}\f$
  double shat(Event const & ev);

  //! Convert an event to a LHEF::HEPEUP
  LHEF::HEPEUP to_HEPEUP(Event const & event, LHEF::HEPRUP *);

}
