/** \file PDG_codes.hh
 *  \brief Contains the Particle IDs of all relevant SM particles.
 *
 *  Large enumeration included which has multiple entries for potential
 *  alternative names of different particles. There are also functions
 *  which can be used to determine if a particle is a parton or if
 *  it is a non-gluon boson.
 *
 *  \authors   Jeppe Andersen, Tuomas Hapola, Marian Heil, Andreas Maier, Jennifer Smillie
 *  \date      2019
 *  \copyright GPLv2 or later
 */

#pragma once

#include <string>

namespace HEJ {

  //! particle ids according to PDG
  namespace pid {
    //! The possible particle identities. We use PDG IDs as standard.
    enum ParticleID{
      d = 1,                          /*!< Down Quark */
      down = d,                       /*!< Down Quark */
      u = 2,                          /*!< Up Quark */
      up = u,                         /*!< Up Quark */
      s = 3,                          /*!< Strange Quark */
      strange = s,                    /*!< Strange Quark */
      c = 4,                          /*!< Charm Quark */
      charm = c,                      /*!< Charm Quark */
      b = 5,                          /*!< Bottom Quark */
      bottom = b,                     /*!< Bottom Quark */
      t = 6,                          /*!< Top Quark */
      top = t,                        /*!< Top Quark */
      e = 11,                         /*!< Electron */
      electron = e,                   /*!< Electron */
      nu_e = 12,                      /*!< Electron Neutrino */
      electron_neutrino = nu_e,       /*!< Electron neutrino */
      mu = 13,                        /*!< Muon */
      muon = mu,                      /*!< Muon */
      nu_mu = 14,                     /*!< Muon Neutrino */
      muon_neutrino = nu_mu,          /*!< Muon Neutrino */
      tau = 15,                       /*!< Tau */
      nu_tau = 16,                    /*!< Tau Neutrino */
      tau_neutrino = nu_tau,          /*!< Tau Neutrino */
      d_bar = -d,                     /*!< Anti-Down Quark */
      u_bar = -u,                     /*!< Anti-Up quark */
      s_bar = -s,                     /*!< Anti-Strange Quark */
      c_bar = -c,                     /*!< Anti-Charm Quark */
      b_bar = -b,                     /*!< Anti-Bottom Quark */
      t_bar = -t,                     /*!< Anti-Top Quark */
      e_bar = -e,                     /*!< Positron */
      positron = e_bar,               /*!< Positron */
      nu_e_bar = -nu_e,               /*!< Anti-Electron Neutrino */
      mu_bar = -mu,                   /*!< Anti-Muon */
      nu_mu_bar = -nu_mu,             /*!< Anti-Muon Neutrino */
      tau_bar = -tau,                 /*!< Anti-Tau */
      nu_tau_bar = -nu_tau,           /*!< Anti-Tau Neutrino */
      gluon = 21,                     /*!< Gluon */
      g = gluon,                      /*!< Gluon */
      photon = 22,                    /*!< Photon */
      gamma = photon,                 /*!< Photon */
      Z = 23,                         /*!< Z Boson */
      Wp = 24,                        /*!< W- Boson */
      Wm = -Wp,                        /*!< W+ Boson */
      h = 25,                         /*!< Higgs Boson */
      Higgs = h,                      /*!< Higgs Boson */
      higgs = h,                      /*!< Higgs Boson */
      p = 2212,                       /*!< Proton */
      proton = p,                     /*!< Proton */
      p_bar = -p,                     /*!< Anti-Proton */
    };

  }

  using ParticleID = pid::ParticleID;

  //! Convert a particle name to the corresponding PDG particle ID
  ParticleID to_ParticleID(std::string const & name);

  /**
   * \brief Function to determine if particle is a parton
   * @param p              PDG ID of particle
   * @returns              true if the particle is a parton, false otherwise
   */
  inline
  constexpr bool is_parton(ParticleID p){
    return p == pid::gluon || std::abs(p) <= pid::top;
  }

  /**
   * \brief function to determine if the particle is a photon, W, Z, or Higgs boson
   * @param id             PDG ID of particle
   * @returns              true if the partice is a A,W,Z, or H, false otherwise
   */
  inline
  constexpr bool is_AWZH_boson(ParticleID id){
    return id == pid::Wm || (id >= pid::photon && id <= pid::Higgs);
  }

}
