/** \file
 *  \brief Interface for pseudorandom number generators
 *
 *  We select our random number generator at runtime according to the
 *  configuration file. This interface guarantees that we can use all
 *  generators in the same way.
 *
 *  \authors   Jeppe Andersen, Tuomas Hapola, Marian Heil, Andreas Maier, Jennifer Smillie
 *  \date      2019
 *  \copyright GPLv2 or later
 */
#pragma once

#include <limits>

namespace HEJ {

  //! Interface for random number generator
  struct RNG {
    //! Generate random number in (0,1]
    virtual double flat() = 0;

    //! Minimum number that can be generated
    virtual unsigned min() const = 0;
    //! Maximum number that can be generated
    virtual unsigned max() const = 0;
    //! Generate random number in [min(), max()]
    virtual unsigned operator()() = 0;

    virtual ~RNG() = default;
  };

  //! Helper struct with default implementations
  struct DefaultRNG : virtual RNG {
    unsigned min() const override {
      return 0u;
    }
    unsigned max() const override {
      return std::numeric_limits<unsigned>::max() - 1;
    }
    unsigned operator()() override {
      return flat()*std::numeric_limits<unsigned int>::max();
    }
  };
}
